#!/usr/bin/env python
# coding: utf-8

# Some L-systems:
#   https://www.robertdickau.com/lsys2d.html
#   https://resources.wolframcloud.com/FunctionRepository/resources/LSystemPlot/

# In[12]:
import math

def get_l_system_points(depth, theta, axiom, ignore, rules, dtheta):
    '''Apply an L-system.'''
    # Build the final script.
    script = axiom
    for i in range(depth):
        new_script = ''
        for ch in script:
            if ch in rules:
                new_script += rules[ch]
            else:
                new_script += ch
        script = new_script
    # print(script)

    # Create the points.
    p1 = (0, 0)
    length = 100
    points = [p1]
    for ch in script:
        if ch in ignore:
            pass
        elif ch == '-':
            theta -= dtheta
        elif ch == '+':
            theta += dtheta
        else:
            x = p1[0] + length * math.cos(theta)
            y = p1[1] + length * math.sin(theta)
            p1 = (x, y)
            points.append(p1)

    # Return the points.
    return points

#%%

def draw_point(canvas, pt, radius, outline, fill):
    canvas.create_oval(
        pt[0] - radius,
        pt[1] - radius,
        pt[0] + radius,
        pt[1] + radius,
        outline=outline,
        fill=fill)

def fit_to_canvas(canvas, curve, margin):
    '''Scale a curve to fit a canvas.'''
    # Get the canvas's dimensions.
    canvas_wid = canvas.winfo_width() - 2 * margin
    canvas_hgt = canvas.winfo_height() - 2 * margin

    # Get the curve's dimensions.
    x1, y1, x2, y2 = canvas.bbox(curve)
    curve_wid = x2 - x1
    curve_hgt = y2 - y1

    # Calculate the scale factor.
    x_scale = (canvas_wid - 2 * margin) / curve_wid
    y_scale = (canvas_hgt - 2 * margin) / curve_hgt
    scale = min(x_scale, y_scale)

    # Scale the curve.
    canvas.scale(curve, 0, 0, scale, scale)

    # Center the curve.
    x1, y1, x2, y2 = canvas.bbox(curve)
    cx = (x1 + x2) / 2
    cy = (y1 + y2) / 2
    xmin = cx - canvas_wid / 2
    xmax = cx + canvas_wid / 2
    ymin = cy - canvas_hgt / 2
    ymax = cy + canvas_hgt / 2
    canvas.config(scrollregion=(xmin, ymin, xmax, ymax))

# In[18]:
import tkinter as tk
from tkinter import ttk
from tkinter import messagebox

class LSystemsApp:
    def __init__(self):
        self.window = tk.Tk()
        self.window.title('l_systems')
        self.window.protocol('WM_DELETE_WINDOW', self.kill_callback)
        self.window.geometry('400x400')

        # Build the UI.
        frame = tk.Frame(self.window)
        frame.pack(side=tk.TOP, anchor=tk.W, padx=3, pady=3)

        label = tk.Label(frame, text='Depth:')
        label.pack(side=tk.LEFT, padx=3)
        self.depth_var = tk.StringVar(value='3')
        entry = tk.Entry(frame, textvariable=self.depth_var, width=3)
        entry.pack(side=tk.LEFT, padx=3)

        label = tk.Label(frame, text='Curve:')
        label.pack(side=tk.LEFT, padx=3)

        # Each L-system includes:
        #   Name: value
        # Where value includes:
        #   Dictionary of rules
        #   Axiom
        #   Dtheta
        #   Max depth
        self.lsystems = {
            '32-Segment Curve': (
                {'F': '-F+F-F-F+F+FF-F+F+FF+F-F-FF+FF-FF+F+F-FF-F-F+FF-F-F+F+F-F+'},
                'F+F+F+F', '', 90, 4),
            'Box Fractal': (
                {'F': 'F-F+F+F-F'},
                'F-F-F-F', '', 90, 7),
            'Dragon Curve': (
                {'X': 'X+YF+', 'Y': '-FX-Y'},
                'FX', '', 90, 20),
            # Hilbert Curve: https://www.fractal.garden/l-system/hilbert-curve
            'Hilbert Curve': (
                {'V': '-WF+VFV+FW-', 'W': '+VF-WFW-FV+'},
                'W', 'VW', 90, 10),
            'Hilbert Curve II': (
                {'X': 'XFYFX+F+YFXFY-F-XFYFX', 'Y': 'YFXFY-F-XFYFX+F+YFXFY'},
                'X', '', 90, 6),
            'Koch Snowflake': (
                {'F': 'F+F--F+F'},
                'F--F--F', '', 60, 7),
            'Peano Curve': (
                {'F': 'F+F-F-F-F+F+F+F-F'},
                'F', '', 90, 6),
            'Peano-Gosper Curve': (
                {'X': 'X+YF++YF-FX--FXFX-YF+', 'Y': '-FX+YFYF++YF+FX--FX-Y'},
                'FX', '', 60, 7),
            'Gosper Boundary': (
                {'A': 'A-A+A'},
                'A-A-A-A-A-A-', '', 60, 7),
            'Quadratic Koch Island': (
                {'F': 'F-F+F+FFF-F-F+F'},
                'F+F+F+F', '', 90, 6),
            'Sierpiński Arrowhead': (
                {'X': 'YF+XF+Y', 'Y': 'XF-YF-X'},
                'YF', '', 60, 10),
            # Sierpiński Curve: https://en.wikipedia.org/wiki/Sierpi%C5%84ski_curve
            'Sierpiński Curve': (
                {'X': 'XF+G+XF--F--XF+G+X'},
                'F--XF--F--XF', 'X', 45, 9),
            # Sierpiński Square Curve: https://en.wikipedia.org/wiki/Sierpi%C5%84ski_curve
            'Sierpiński Square Curve': (
                {'X': 'XF-F+F-XF+F+XF-F+F-X'},
                'F+XF+F+XF', '', 90, 9),
            'Siepiński Sieve': (
                {'F': 'FF', 'X': '--FXF++FXF++FXF--'},
                'FXF--FF--FF', '', 60, 10),
        }

        curve_names = list(self.lsystems.keys())
        self.curve_combo = ttk.Combobox(frame, values=curve_names,
                                        state='readonly')
        self.curve_combo.pack(side=tk.LEFT, padx=3)
        self.curve_combo.current(0)

        button = tk.Button(frame, text='Draw', width=10, command=self.draw)
        button.pack(side=tk.LEFT, padx=10)

        self.canvas = tk.Canvas(self.window, bg='white',
            borderwidth=2, relief=tk.SUNKEN)
        self.canvas.pack(padx=10, pady=(0, 10),
            side=tk.BOTTOM, fill=tk.BOTH, expand=True)

        # If the user presses Enter, draw.
        entry.bind('<Return>', lambda event: self.draw())

        # Display the window.
        entry.focus_force()
        self.window.mainloop()

    def kill_callback(self):
        # Destroy the tkinter window.
        self.window.destroy()

    def draw(self):
        # Delete any previous drawing objects.
        self.canvas.delete(tk.ALL)

        # Get the desired depth.
        try:
            depth = int(self.depth_var.get())
        except:
            depth = -1

        # See which curve is selected.
        curve_name = self.curve_combo.get()
        lsystem = self.lsystems[curve_name]
        rules, axiom, ignore, angle, max_depth = lsystem
        dtheta = math.radians(angle)
        theta = 0

        # See if it's the Sierpiński Arrowhead.
        if curve_name == 'Sierpiński Arrowhead':
            # Draw odd-numbered depths backward so they all point down.
            if depth % 2 == 1:
                theta = math.radians(180)

        # Validate the desired depth.
        if depth < 0 or depth > max_depth:
            messagebox.showinfo('Depth Error',
                f'{curve_name} depth must be an integer between 0 and {max_depth}.')
            return

        # Generate the curve.
        points = get_l_system_points(depth, theta, axiom, ignore, rules, dtheta)

        # Draw the curve.
        curve = self.canvas.create_line(points, width=0, fill='black')

        # Scale to fit.
        fit_to_canvas(self.canvas, curve, 10)


# In[19]:
LSystemsApp()
print('Done')
